// Emacs will be in -*- Mode: c++ -*-
#ifndef _BiCubicSpline_h_
#define _BiCubicSpline_h_

#include <iomanip>
#include <vector>

#include "CubicSpline.h"

using std::vector;

/*!
  \class BiCubicSpline
  \brief Bicubic spline interpolation class
  \author Nicolas Di Csar 
  \version 0.91
  \date April 2001
  \remark use CubicSpline to perform bicubic spline interpolation
  \sa CubicSpline
*/
template <class T>
class BiCubicSpline
{
private:
  //! (x,y) data points.
  vectorView<T> x_, y_;
  //! z matrix of values at (x,y)
  matrixView<T> z_;

  //! vector of cubic spline interpolation
  vector< CubicSpline<T> >  bcs_;

  //! to know what is the first interpolation direction 
  bool is_y_coef_; 
  //! to know if the interpolation coefficients are already computed
  bool is_coef_computed_;
public:
  BiCubicSpline();
  //! Constructor to initialize x,y,z data
  BiCubicSpline(T * vx, size_t vx_sz, 
		T * vy, size_t vy_sz,
		T* mz);
  //! Destructor
  ~BiCubicSpline() {}

  //! Set Data if default constructor is called
  void setData(T * vx, size_t vx_sz, 
		T * vy, size_t vy_sz,
		T* mz);
  
  //! compute spline coefficents in x direction
  void Set_X_Spline_Coef();
  //! compute spline coefficents in y direction
  void Set_Y_Spline_Coef();

  //! Evaluate cubic spline interpolation at (xx,yy).
  T operator ()(T xx, T yy);

  //! Evaluate 1st partial derivative in x of cubic spline interpolation at (xx,yy).
  T dx(T xx, T yy);
  //! Evaluate 2nd partial derivative in x of cubic spline interpolation at (xx,yy).
  T dxx(T xx, T yy);

  //! Evaluate 1st partial derivative in y of cubic spline interpolation at (xx,yy).
  T dy(T xx, T yy);
  //! Evaluate 2nd partial derivative in y of cubic spline interpolation at (xx,yy).
  T dyy(T xx, T yy);
protected:
  //! Perform interpolation using y-direction coefficient
  T y_interpolation(T xx, T yy);
};

template <class T> 
BiCubicSpline<T>::BiCubicSpline() : x_(), y_(), z_(), is_y_coef_(1), is_coef_computed_(0)
{}

template <class T> 
BiCubicSpline<T>::BiCubicSpline(T * vx, size_t vx_sz,
				T * vy, size_t vy_sz,
				T* mz)
  : x_(vx_sz,vx), y_(vy_sz,vy), z_(vx_sz,vy_sz,mz), is_y_coef_(1), is_coef_computed_(0)
{}


template <class T> 
void BiCubicSpline<T>::setData(T * vx, size_t vx_sz, 
		T * vy, size_t vy_sz,
		T* mz)
{
	x_ = vectorView<T>(vx_sz,vx);
	y_ = vectorView<T>(vy_sz,vy);
	z_ = matrixView<T>(vx_sz,vy_sz,mz);

	is_coef_computed_ = 0;
}


template <class T>
void BiCubicSpline<T>::Set_X_Spline_Coef()
{
  if ( (!is_coef_computed_) || (is_y_coef_) ) {

    int i, j, xsz = x_.size(), ysz = y_.size();

    if (!is_coef_computed_) {
      bcs_ = vector< CubicSpline<T> >(ysz);
    }
    else {
      bcs_ = vector< CubicSpline<T> >(ysz);
    }

    vector<T> vx(xsz);
    
    for (i=0; i<ysz; ++i){
      
      for (j=0; j<xsz; ++j)
	vx[j] = z_(j,i);
	
	bcs_[i].setData(x_.begin(),x_.size(), vx.begin(), vx.size());
    }

    is_y_coef_ = 0;
    is_coef_computed_ = 1;
  }
}

template <class T>
void BiCubicSpline<T>::Set_Y_Spline_Coef()
{
  if ( (!is_coef_computed_) || (!is_y_coef_) ) {

    int i, xsz = x_.size();
      
    if (!is_coef_computed_) {
      bcs_ = vector< CubicSpline<T> >(xsz) ;
    }
    else {
      bcs_ = vector< CubicSpline<T> >(xsz);
    }
      
    for (i=0; i<xsz; ++i) {
      bcs_[i].setData(y_.begin(),y_.size(), z_[i].begin(),z_[i].size());
    }

    is_y_coef_ = 1;
    is_coef_computed_ = 1;
  }
}

template <class T>
T BiCubicSpline<T>::operator ()(T xx, T yy)
{
  return y_interpolation(xx, yy);
}

template <class T>
T BiCubicSpline<T>::y_interpolation(T xx, T yy)
{
  Set_Y_Spline_Coef();

  int i, sz = x_.size();
  vector<T> cs_values(sz);

  for (i=0; i<sz; ++i)
    cs_values[i] = bcs_[i](yy);
  
  CubicSpline<T> local_cs(x_.begin(),x_.size(), 
			  cs_values.begin(),cs_values.size());
  
  return local_cs(xx);
}

template <class T>
T BiCubicSpline<T>::dx(T xx, T yy)
{
  Set_X_Spline_Coef();
  
  int i, sz = y_.size();
  vector<T> dx_cs_values(sz);
  for (i=0; i<sz; ++i)
    dx_cs_values[i] = bcs_[i].dx(xx);

  CubicSpline<T> local_cs(y_.begin(),y_.size(), 
			  dx_cs_values.begin(),dx_cs_values.size());

  return local_cs(yy);
}

template <class T>
T BiCubicSpline<T>::dxx(T xx, T yy)
{
  Set_X_Spline_Coef();

  int i, sz = y_.size();
  vector<T> dxx_cs_values(sz);
  for (i=0; i<sz; ++i)
    dxx_cs_values[i] = bcs_[i].dxx(xx);

  CubicSpline<T> local_cs(y_.begin(),y_.size(), 
			  dxx_cs_values.begin(),dxx_cs_values.size());

  return local_cs(yy);
}

template <class T>
T BiCubicSpline<T>::dy(T xx, T yy)
{
  Set_Y_Spline_Coef();

  int i, sz = x_.size();
  vector<T> dx_cs_values(sz);
  for (i=0; i<sz; ++i)
    dx_cs_values[i] = bcs_[i].dx(yy);

  CubicSpline<T> local_cs(x_.begin(), x_.size(),
			  dx_cs_values.begin(), dx_cs_values.size());

  return local_cs(xx);
}

template <class T>
T BiCubicSpline<T>::dyy(T xx, T yy)
{
  Set_Y_Spline_Coef();

  int i, sz = x_.size();
  vector<T> dxx_cs_values(sz);
  for (i=0; i<sz; ++i)
    dxx_cs_values[i] = bcs_[i].dxx(yy);

  CubicSpline<T> local_cs(x_.begin(),x_.size(), 
			  dxx_cs_values.begin(),dxx_cs_values.size());

  return local_cs(xx);
}

#endif
