// Emacs will be in -*- Mode: c++ -*-
#ifndef _CubicSpline_h_
#define _CubicSpline_h_

#include <ctype.h>
#include <string.h>
#include <vector>

#include "view.h"
#include "dgtsln.h"
#include "error.h"




/*!
  \class CubicSpline
  \brief Cubic spline interpolation class
  \author Nicolas Di Csar 
  \version 0.91
  \date April 2001
  The MATLAB De Boor algorithm is used
  \remark subroutine dgtsln from LINPACK, converted to C from Fortran (should be included)
*/
template <class T>
class CubicSpline
{
private:
  //! (x,y) data points.
  vectorView<T> x_, y_;

  //! 1,2,3-rd Cubic spline coefficients.
  std::vector<T> a1, a2, a3;
  
public:
  /*!
    \brief Empty constructor to allow array of cubic spline 
    \remark useful for bicubic spline
    \warning Call setData before doing anything
    \sa setData
  */
  CubicSpline();
  
  //! Compute the cubic spline coefficients
  CubicSpline(T * vx, size_t vx_sz, T * vy, size_t vy_sz);
  
  //! Copy constructor
  CubicSpline(const CubicSpline &cub);
  
  //! Destructor
  virtual ~CubicSpline() {}
  
  //! Assignment operator.
  CubicSpline& operator=(const CubicSpline& cub);
  
  //! Set Data if default constructor is called
  void setData(T * vx, size_t vx_sz, T * vy, size_t vy_sz);

  //! return ith value of first coefficient 
  T getA1(int i) const;
  //! return ith value of 2nd coefficient
  T getA2(int i) const;
  //! return ith value of 3rd coefficient
  T getA3(int i) const;
  
  //! Compute the cubic spline with MATLAB De Boor algorithm
  void cubspline();      

  //! search position of xx in the array x_
  size_t search_pos(T xx) const;
  
	//! Extrapolation
  T extrapolation(T xx) const;

  //! Evaluate cubic spline interpolation at xx.
  T operator ()(T xx) const;

  //! Evaluate the first derivative of cubic spline interpolation at xx.
  T dx(T xx) const;
  
  //! Evaluate the 2nd derivative of cubic spline interpolation at xx.
  T dxx(T xx) const;
  
  const vectorView<T>& x() const { return x_;}
  const vectorView<T>& y() const { return y_;}
  
};

template <class T>
inline CubicSpline<T>::CubicSpline() : x_(), y_() {}


template <class T>
inline CubicSpline<T>::CubicSpline(T * vx, size_t vx_sz, T * vy, size_t vy_sz)
  : x_(vx_sz,vx), y_(vy_sz,vy), a1(vx_sz), a2(vx_sz), a3(vx_sz)
{
  cubspline();// Compute the cubic spline coefficients. 
}

template <class T>
inline void CubicSpline<T>::setData(T * vx, size_t vx_sz, T * vy, size_t vy_sz)
{
  x_ = vectorView<T>(vx_sz,vx); y_ = vectorView<T>(vy_sz,vy);
  
  a1 = vector<T>(vx_sz);
  a2 = vector<T>(vx_sz);
  a3 = vector<T>(vx_sz);

  cubspline();// Compute the cubic spline coefficients. 
}


template <class T>
inline CubicSpline<T>::CubicSpline(const CubicSpline<T> &cub)
  : x_(cub.x_), y_(cub.y_), a1(cub.a1), a2(cub.a2), a3(cub.a3) {}

template <class T>
inline T CubicSpline<T>::getA1(int i) const
{
    return a1[i];
}

template <class T>
inline T CubicSpline<T>::getA2(int i) const
{
    return a2[i];
}

template <class T>
inline T CubicSpline<T>::getA3(int i) const
{
    return a3[i];
}

template <class T>
void CubicSpline<T>::cubspline(void)
{
  int i;
  int size = a1.size();
  int size1 = size - 1;

  // Check for data consistency.
  for( i = 0 ; i < size1 ; i++)
    if( x_[i] >= x_[i+1] )
      throw Error("x data not in ascending order");

  // Fill the tridiagonal matrix except for first and last rows.
  vector<T>  b(2*size); // b is the rhs vector.
    
  for ( i = 1 ; i < size1  ; i++) {
    T dxip = x_[i+1] - x_[i], divdifim = (y_[i] - y_[i-1])/(x_[i] - x_[i-1]);
    T dxim = x_[i] - x_[i-1], divdifip = (y_[i+1] - y_[i])/(x_[i+1] - x_[i]);
    
    a3[i] = 2.0 * ( dxip + dxim );  // a3 is the      diagonal
    a1[i] = dxip;                   // a1 is the   subdiagonal
    a2[i] = dxim;                   // a2 is the superdiagonal
    
    b[2*i] = 3.0 * ( dxip * divdifim + dxim * divdifip ); 
  }

  // Evaluate cubic spline coefficient a2 depending on type.
  int ldb = 2;        // dgtsln variables.
  int nrhs, info;

  T dx0 = x_[1] - x_[0], dxsize1 = x_[size1]-x_[size1-1], 
    dxsize2 = x_[size1-1]-x_[size1-2], dx1 = x_[2] - x_[1];
  T dy0 = y_[1] - y_[0], dysize1 = y_[size1]-y_[size1-1], 
    dysize2 = y_[size1-1]-y_[size1-2], dy1 = y_[2] - y_[1];
  T x20 = x_[2] - x_[0], xsize1 = x_[size1]-x_[size1-2];
  
  a1[0]  = 0;
  a3[0]  = dx0;
  a2[0]  = x20;
  b[0]   = ((dx0+2.*x20)*dx1*dy0/dx0 + dx0*dx0*dy1/dx1)/x20;
  a3[size1]  = dxsize1;
  a1[size1]  = xsize1;
  b[2*size1] = ( dxsize1*dxsize1*dysize2/dxsize2 + (2.*xsize1+dxsize1)*dxsize2*dysize1/dxsize1)/xsize1;
  
  // Solve tridiagonal system.
  nrhs = 1;
  dgtsln(nrhs, ldb, a1, a3, a2, b, info);
  
  // Evaluate a1,a2 and a3.
  for( i = 0 ; i < size1 ; i++) {
    T dxi = x_[i+1]-x_[i], dyi = y_[i+1]-y_[i];
    a3[i] = ( b[2*i]+b[2*(i+1)]-2.*dyi/dxi)/dxi;
    a2[i] = (dyi/dxi-b[2*i])/dxi - a3[i] ;
    a3[i] /= dxi;
    a1[i] = b[2*i];
  }
    
}


template <class T>
size_t CubicSpline<T>::search_pos(T xx) const
{
  int size = x_.size();

  // Find the relevant interval.
  int i;
  for (i = 1 ; (xx >= x_[i]) && (i < size)  ; i++);
  i--;

  return i;
}


template <class T>
T CubicSpline<T>::extrapolation(T xx) const
{

  

  int size = a1.size(),i=0;
  const int size1 = size-1;

  // Check that xx is outside valid range.
  if ( (xx >= x_[0]) && (xx <= x_[size1])) 
	{
		cerr << x_[0] << " < " << xx << " < " << x_[size1] << endl;
		throw Error("Use interpolation instead of extrapolation");
	}
  
  if ( xx > x_[size1]) i = size1;

	// Evaluate the interpolant.
  T xd = xx - x_[i];

  return (y_[i] + xd * (a1[i] + xd * (a2[i] + xd * a3[i])));

}

// Evaluate cubic spline interpolation at x.
template <class T>
T CubicSpline<T>::operator ()(T xx) const
{
  int size = a1.size();
  const int size1 = size-1;

  if (xx == x_[size1])
    return y_[size1];

  // Check that xx falls within valid range.
  if ( (xx < x_[0]) || (xx > x_[size1])) 
	{
		cerr << x_[0] << " < " << xx << " < " << x_[size1] << endl;
		throw Error("xx out of range, use extrapolation");
	}
  
  // Find the relevant interval.
  int i = search_pos(xx);
  
  // Evaluate the interpolant.
  T xd = xx - x_[i];

  return (y_[i] + xd * (a1[i] + xd * (a2[i] + xd * a3[i])));

}

//Evaluate the first derivative of cubic spline interpolation at xx.
template <class T>
T CubicSpline<T>::dx(T xx) const
{
  int size = a1.size();
  const int size1 = size-1;

  // Check that xx falls within valid range.
  if ( (xx < x_[0]) || (xx > x_[size1]))
	{
		cerr << x_[0] << " < " << xx << " < " << x_[size1] << endl;
		throw Error("xx out of range");
	}
  
  T xd;
  if (xx == x_[size1]) {
    xd = xx - x_[size1-1];

    return (a1[size1-1] + xd * (2.*a2[size1-1] + xd * 3. * a3[size1-1]));
  }

  // Find the relevant interval.
  int i = search_pos(xx);

  // Evaluate the interpolant.
  xd = xx - x_[i];
  
  return (a1[i] + xd * (2.*a2[i] + xd * 3. * a3[i]));
}


//Evaluate the 2nd derivative of cubic spline interpolation at xx.
template <class T>
T CubicSpline<T>::dxx(T xx) const
{
  int size = a1.size();
  const int size1 = size-1;

  // Check that xx falls within valid range.
  if ( (xx < x_[0]) || (xx > x_[size1]))
	{
		cerr << x_[0] << " < " << xx << " < " << x_[size1] << endl;
		throw Error("xx out of range");
	}

  T xd;
  if (xx == x_[size1]) {
    xd = xx - x_[size1-1];
    return (2.*a2[size1-1] + xd * 6. * a3[size1-1]);
  }

  // Find the relevant interval.
  int i = search_pos(xx);

  // Evaluate the interpolant.
  xd = xx - x_[i];
  
  return (2.*a2[i] + xd * 6. * a3[i]);
}

// Assignment operator.
template <class T>
CubicSpline<T>& CubicSpline<T>::operator=(const CubicSpline<T>& cub)
{
    if(this != &cub) { // Do nothing if cub = cub
      x_ = cub.x_;
      y_ = cub.y_;
      a1 = cub.a1;
      a2 = cub.a2;
      a3 = cub.a3;
    }

    return *this;
}

#endif
