// bicubic_test.cpp  Example of using the CubicSpline class.

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cmath>
#include <cstdlib>

//#define DEBUG

#include "BiCubicSpline.h"

#include <vector>

using std::vector;

double func(double x, double y)
{
  return 3*x + 3*y + 2*x*x + 2*y*y + x*x*x + y*y*y;
}

double dyfunc(double x, double y)
{
  return 3. + 4.*y + 3.*y*y;
}

double dxfunc(double x, double y)
{
  return 3. + 4.*x + 3.*x*x;
}

double dyyfunc(double x, double y)
{
  return 4.+6.*y;
}

double dxxfunc(double x, double y)
{
  return 4.+6.*x;
}


class Fxy {
public:
  typedef double (*ptr_fxy)(double,double);
protected:
  ptr_fxy fxy_, dxfxy_, dxxfxy_, dyfxy_, dyyfxy_;
public:
  Fxy(ptr_fxy fxy, 
      ptr_fxy dxfxy, ptr_fxy dxxfxy, 
      ptr_fxy dyfxy, ptr_fxy dyyfxy)
    : fxy_(fxy), dxfxy_(dxfxy), dxxfxy_(dxxfxy), dyfxy_(dyfxy), dyyfxy_(dyyfxy) {}
  ~Fxy() {}

  double operator() (double xx, double yy) { return (*fxy_)(xx,yy);}
  double dx(double xx, double yy) { return (*dxfxy_)(xx,yy);}
  double dxx(double xx, double yy) { return (*dxxfxy_)(xx,yy);}
  double dy(double xx, double yy) { return (*dyfxy_)(xx,yy);}
  double dyy(double xx, double yy) { return (*dyyfxy_)(xx,yy);}
};

void bicubic_spline_test(Fxy& func)
{
  size_t i,j;

  const size_t nx = 6;//11            // Number of data points.
  const size_t ny = 6; 
  
  const double xmin    = 0.0;  // Range of independent variable.
  const double xmax    = (double)nx-1;  
  const double xlength = xmax-xmin;
  const double hx      = xlength/double(nx-1);
  
  const double ymin    = 0.0;  // Range of independent variable.
  const double ymax    = (double)ny-1;  
  const double ylength = ymax-ymin;
  const double hy      = ylength/double(ny-1);
  
  try{
    // Construct data points (x,y,z) for which bicubic spline is to be computed.
    vector<double> x(nx,0.), y(ny,0.);
    vector<double> z(nx*ny,0);

    for (i = 0 ; i < nx ; i++) {
      x[i] = (xmin+hx*double(i));
    }

    for (j = 0 ; j < ny ; j++) {
      y[j] = (ymin+hy*double(j));
    }

    for (i = 0 ; i < nx ; i++) {
      for (j = 0 ; j < ny ; j++) {
		z[i*ny+j] = func(x[i],y[j]);
      }
    }
    
    // Construct cubic splines.
    // Natural bicubic spline : Matlab method
    BiCubicSpline<double> spline(x.begin(), x.size(), y.begin(), y.size(), z.begin());               
    
    // Sample all the functions at a fine grid.
    const size_t nnx = 21;
    const size_t nny = 11;
    vector<double> xx(nnx,0.), yy(nny,0.);
    vector<double> zz(nnx*nny), zn(nnx*nny);
  
    const double hhx =  xlength/double(nnx-1);
    const double hhy =  ylength/double(nny-1);
    
    double fxy,dxfxy,dxxfxy,dyfxy,dyyfxy;
    double efxy,edxfxy,edxxfxy,edyfxy,edyyfxy;
    efxy = edxfxy = edxxfxy = edyfxy = edyyfxy = 0.;
    for (i = 0 ; i < nnx ; ++i) {
      double xi = xmin + hhx*double(i);
      for (j = 0 ; j < nny ; ++j) {
	double yj = ymin + hhy*double(j);

	xx[i]  = xi;
	yy[j]  = yj;
	zz[i*nny+j]  = func(xi,yj);
	
	zn[i*nny+j]  = spline(xi,yj);

	fxy = zn[i*nny+j] - zz[i*nny+j];
	dxfxy = spline.dx(xi,yj) - func.dx(xi,yj);
	dxxfxy = spline.dxx(xi,yj) - func.dxx(xi,yj);
	dyfxy = spline.dy(xi,yj) - func.dy(xi,yj);
	dyyfxy = spline.dyy(xi,yj) - func.dyy(xi,yj);
/*
  cout << "zn[i][j] = " << setw(14) << fxy
  << ", dy(zn[i][j]) = " << setw(14) << dyfxy
  << ", dyy(zn[i][j]) = " << setw(14) << dyyfxy
  << ", dx(zn[i][j]) = " << setw(14) << dxfxy
  << ", dxx(zn[i][j]) = " << setw(14) << dxxfxy
  << endl;
*/
	efxy += fxy*fxy;
	edxfxy += dxfxy*dxfxy;
	edxxfxy += dxxfxy*dxxfxy;
	edyfxy += dyfxy*dyfxy;
	edyyfxy += dyyfxy*dyyfxy;
      }
    }
  
    cout << "fxy error    = " << sqrt(efxy) << endl
	     << "dxfxy error  = " << sqrt(edxfxy) << endl
	     << "dxxfxy error = " << sqrt(edxxfxy) << endl
	     << "dyfxy error  = " << sqrt(edyfxy) << endl
	     << "dyyfxy error = " << sqrt(edyyfxy) << endl;

  }
  catch(...) {
  }

} 


int main()
{
  Fxy fun_xy(&func, 
	     &dxfunc, &dxxfunc, 
	     &dyfunc, &dyyfunc);

  cout << endl 
	   << "********* 2D Polynomial function of degree 3 ********" << endl;
  cout << "*    f(x,y) = 3x + 3y + 2x^2 + 2y^2 + x^3 + y^3     *" << endl
	   << "*****************************************************" << endl;
  bicubic_spline_test(fun_xy);
  cout << endl << endl;

  return 0;
}

// Emacs settings.
// Local Variables:
// compile-command:"make -k bicubic_test"
