#include <iostream>
#include <iomanip>
#include <fstream>
#include <cmath>
#include <vector>


#include <CubicSpline.h>
#include <ql/Math/cubicspline.hpp>

using std::vector;

class Fx {
public:
  typedef double (*ptr_fx)(double);
protected:
  ptr_fx fx_, dxfx_, dxxfx_;
public:
  Fx(ptr_fx fx, ptr_fx dxfx, ptr_fx dxxfx)
    : fx_(fx), dxfx_(dxfx), dxxfx_(dxxfx) {}
  ~Fx() {}
  double operator() (double xx) { return (*fx_)(xx);}
  double dx(double xx) { return (*dxfx_)(xx);}
  double dxx(double xx) { return (*dxxfx_)(xx);}
};

double func(double x);
double dxfunc(double x);
double dxxfunc(double x);

double func2(double x);
double dxfunc2(double x);
double dxxfunc2(double x);

double func3(double x);
double dxfunc3(double x);
double dxxfunc3(double x);

void cubic_spline_test(Fx& func);
void cs_quantlib_test(Fx& func);

int main()
{
  cout << endl 
	   << "********* Polynomial function of degree 1 ********" << endl;
  cout << "*               f(x) = 2 + x                     *" << endl
	   << "**************************************************" << endl;
  Fx fun_x(&func, &dxfunc, &dxxfunc);

  cout << "Natural cubic spline : Matlab Method";
  cubic_spline_test(fun_x);
  
  cout << endl << "Quantlib cubic spline";
  cs_quantlib_test(fun_x);

  cout << endl 
	   << "********* Polynomial function of degree 2 ********" << endl;
  cout << "*           f(x) =  3 + 2x + x^2                 *" << endl
	   << "**************************************************" << endl;
  Fx fun_x2(&func2, &dxfunc2, &dxxfunc2);

  cout << "Natural cubic spline : Matlab Method";
  cubic_spline_test(fun_x2);
  
  cout << endl << "Quantlib cubic spline";
  cs_quantlib_test(fun_x2);

  cout << endl 
	   << "********* Polynomial function of degree 3 ********" << endl;
  cout << "*           f(x) = 4 + 3x + 2x^2 + x^3           *" << endl
	   << "**************************************************" << endl;
  Fx fun_x3(&func3, &dxfunc3, &dxxfunc3);

  cout << "Natural cubic spline : Matlab Method";
  cubic_spline_test(fun_x3);  
  cout << endl << "Quantlib cubic spline";
  cs_quantlib_test(fun_x3);



  return 0;
}


double func(double x)
{
  return 2. + x;
}

double dxfunc(double x)
{
  return 1.;
}

double dxxfunc(double x)
{
  return 0.;
}


double func2(double x)
{
  return 3. + 2.*x + x*x;
}

double dxfunc2(double x)
{
  return 2. + 2.*x;
}

double dxxfunc2(double x)
{
  return 2.;
}


double func3(double x)
{
  return 4. + 3.*x + 2.*x*x + x*x*x;
}

double dxfunc3(double x)
{
  return 3 + 4.*x + 3.*x*x;
}

double dxxfunc3(double x)
{
  return 4.+6.*x;
}


void cs_quantlib_test(Fx& func)
{
  int i;

  const int n = 11;            // Number of data points.
  const double xmin    = 0.0;  // Range of independent variable.
  const double xmax    = (double)n-1;  
  const double xlength = xmax-xmin;
  const double h       = xlength/double(n-1);

  // Construct data points (x,y) for which cubic spline is to be computed.
  vector<double> x(n), y(n);

  for (i = 0 ; i < n ; i++) {
    x[i] = (xmin+h*double(i));
    y[i] = func(x[i]);
  }

  // Construct cubic splines.
  try{
	  typedef double* myIterator; 
    // Natural cubic spline
	  QuantLib::Math::CubicSpline<myIterator,myIterator> spline(x.begin(),x.end(), y.begin());               

    // the functions at a fine grid.
    const int nn = 41;
    vector<double> xx(nn), 
      yy(nn),             
      ym(nn);            
    const double hh =  xlength/double(nn-1);

    double fx;
    double efx;
    efx = 0.;
    for (i = 0 ; i < nn ; i++) {
      double xi = xmin + hh*double(i);
      xx[i]  = xi;
      yy[i]  = func(xi);
      ym[i]  = spline(xi);
      
      fx = ym[i] - yy[i];

      efx += fx*fx;
    }
    cout << endl << endl;

    cout << "fx error   = " << sqrt(efx) << endl;
	
  } catch (...) {
      
  }

}



void cubic_spline_test(Fx& func)
{
  int i;

  const int n = 11;            // Number of data points.
  const double xmin    = 0.0;  // Range of independent variable.
  const double xmax    = (double)n-1;  
  const double xlength = xmax-xmin;
  const double h       = xlength/double(n-1);

  // Construct data points (x,y) for which cubic spline is to be computed.
  vector<double> x(n), y(n);

  for (i = 0 ; i < n ; i++) {
    x[i] = (xmin+h*double(i));
    y[i] = func(x[i]);
  }

  // Construct cubic splines.
  try{
    // Natural cubic spline
    CubicSpline<double> spline(x.begin(),x.size(), y.begin(),y.size());               

    // the functions at a fine grid.
    const int nn = 41;
    vector<double> xx(nn), 
      yy(nn),             
      ym(nn);            
    const double hh =  xlength/double(nn-1);

    double fx, dxfx, dxxfx;
    double efx, edxfx, edxxfx;
    efx = edxfx = edxxfx = 0.;
    for (i = 0 ; i < nn ; i++) {
      double xi = xmin + hh*double(i);
      xx[i]  = xi;
      yy[i]  = func(xi);
      ym[i]  = spline(xi);
      
      fx = ym[i] - yy[i];
      dxfx = spline.dx(xi) - func.dx(xi);
      dxxfx = spline.dxx(xi) - func.dxx(xi);

      efx += fx*fx;
      edxfx += dxfx*dxfx;
      edxxfx += dxxfx*dxxfx;
    }
    cout << endl << endl;

    cout << "erreur fx    = " << sqrt(efx) << endl
	 << "erreur dxfx  = " << sqrt(edxfx) << endl
	 << "erreur dxxfx = " << sqrt(edxxfx) << endl;

  }  catch (...) {
      
  }

}

// Emacs settings.
// Local Variables:
// compile-command:"make -k cubic_test"
