// Emacs will be in -*- Mode: c++ -*-
#ifndef __dgtsln_h__
#define __dgtsln_h__

#include <math.h>  


template <class V>
void dgtsln(int &nrhs, int &ldb, V& c, V& d, V& e, V& b, int &info);

template <class T>
void dgtsln(int &n, int &nrhs, int &ldb, T *c, T *d, T *e, T *b, int &info);

template <class V>
inline void dgtsln(int &nrhs, int &ldb, V& c, V& d, V& e, V& b, int &info)
{
  int n = c.size();
  dgtsln(n, nrhs, ldb, c.begin(), d.begin(), e.begin(), b.begin(), info);
}

/*!
 \brief Tridiagonal solver
 \remark Converted from dgtsln.f by f2c

 dgtsln given a general tridiagonal matrix and a right hand
 side will find the solution.
 Ueda: revised from dgtsl to solve for ldb number of rhs, each
 a row of b.
 on entry
 n       integer
         is the order of the tridiagonal matrix.
 
 nrhs    integer
         is the number of rhs to solve for.
 ldb     integer
         is the leading order of b.

 c       double precision(n)
         is the subdiagonal of the tridiagonal matrix.
         c(2) through c(n) should contain the subdiagonal.
         on output c is destroyed.
 d       double precision(n)
         is the diagonal of the tridiagonal matrix.
         on output d is destroyed.
 e       double precision(n)
         is the superdiagonal of the tridiagonal matrix.
         e(1) through e(n-1) should contain the superdiagonal.
         on output e is destroyed.

 b       double precision(n*ldb)
         is the right hand side vectors, each in one row.
 on return
 b       is the solution vectors.
 info    integer
         = 0 normal value.
         = k if the k-th element of the diagonal becomes
             exactly zero.  the subroutine returns when
             this is detected.
 linpack. this version dated 08/14/78 .
 jack dongarra, argonne national laboratory.
*/
template <class T>
void dgtsln(int &n, int &nrhs, int &ldb,
	     T *c, T *d, T *e, T *b, int &info)
{
  int k, kb, ku, kp1, nm1, nm2;;
  T t;
  
  // kludge the Fortran fencepost.
  int b_dim1, b_offset;
  b_dim1 = ldb;
  b_offset = b_dim1 + 1;
  b -= b_offset;
  --c;
  --d;
  --e;
	
  // begin block permitting 
  info = 0;
  c[1] = d[1];
  nm1 = n - 1;
  if (nm1 >= 1)
    {
      d[1] = e[1];
      e[1] = 0.0;
      e[n] = 0.0;

      for (k = 1; k <= nm1 ; ++k) {
	kp1 = k + 1;

	// find the largest of the two rows 
	if ( fabs(c[kp1]) >=  fabs(c[k]))
	  {
				// interchange row 
	    t = c[kp1];
	    c[kp1] = c[k];
	    c[k] = t;
	    t = d[kp1];
	    d[kp1] = d[k];
	    d[k] = t;
	    t = e[kp1];
	    e[kp1] = e[k];
	    e[k] = t;
	    for (ku = 1; ku <= nrhs; ++ku) {
	      t = b[ku + kp1 * b_dim1];
	      b[ku + kp1 * b_dim1]
		= b[ku + k * b_dim1];
	      b[ku + k * b_dim1] = t;
	    }
	  }
		
	// zero elements 
	if (c[k] == 0.0)
	  {
	    info = k;
	    return;
	  }
		
	t = -c[kp1] / c[k];
	c[kp1] = d[kp1] + t * d[k];
	d[kp1] = e[kp1] + t * e[k];
	e[kp1] = 0.0;
	for (ku = 1; ku <= nrhs; ++ku) 
	  b[ku + kp1 * b_dim1] += t * b[ku + k * b_dim1];

      }
    }
	
  if (c[n] == 0.)
    {
      info = n;
      return;
    }

  // back solve 
  nm2 = n - 2;
  for (ku = 1; ku <= nrhs; ++ku) 
    b[ku + n * b_dim1] /= c[n];
  if (n == 1)
    return;

  for (ku = 1; ku <= nrhs; ++ku) 
    b[ku+nm1*b_dim1] = (b[ku+nm1*b_dim1]
			- d[nm1] * b[ku+n*b_dim1]) / c[nm1];
  if (nm2 < 1)
    return;

  for (kb = 1; kb <= nm2; ++kb)
    {
      k = nm2 - kb + 1;
      for (ku = 1; ku <= nrhs; ++ku) 
	b[ku+k*b_dim1] = (b[ku+k*b_dim1]
			  - d[k] * b[ku+(k+1)* b_dim1]
			  - e[k] * b[ku+(k+2)* b_dim1]) / c[k];
    }
  return;
}



#endif
