// Emacs will be in -*- Mode: c++ -*-
#ifndef _view_h_
#define _view_h_

#include <iostream>


/*!
  Vector wrapper class for already allocated vector.
  It allows bounds checking in operator[] but doesn't manage
  memory allocation/desallocation. It simply gives a way to
  use many vector class. 
 */
template <class T>
class vectorView {
  //! Vector size
  size_t sz_;
  //! Pointer to vector data
  T* data_;
public:
  //! Default constructor
  inline explicit vectorView() : sz_(0), data_((T*)0) {}
  //! Main constructor to set member data
  inline vectorView(size_t sz, T* data)
    : sz_(sz), data_(data) {}
  //! Copy constructor
  inline vectorView(const vectorView& vv)
    : sz_(vv.sz_), data_(vv.data_) {}
  //! Destructor
  inline ~vectorView() {} // nothing to delete because it's wrapper
  
  //! Assignement operator
  inline vectorView& operator= (const vectorView& vv)
  {
    sz_ = vv.sz_; data_ = vv.data_;
    return *this;
  }
  
  //! return the ith element of the underlying vector
  inline const T& operator [] (size_t i) const
  {
#ifdef CHECK_SIZE
    if ( data_ == 0 ) throw Error("vectorView<>::operator[], empty array");
    
    if ( !( (i >= 0) && (i < sz_) ) ) {
      cerr << "i = " << i << ", size = " << sz_ << endl;
      throw Error("vectorView<>::operator[], index out of bound");
    }
#endif
    
    return *(data_ + i);
  }
  
  //! return size
  inline size_t size() const  { return sz_;}
  //! return pointer to data of the underlying vector
  inline T*     begin() const { return data_;}
};


/*!
  Matrix wrapper for already allocated matrix.
  It allows bounds checking in operator(,) but doesn't manage
  memory allocation/desallocation. It simply gives a way to
  use many matrix class. Assume that matrix is store in a single
  vector
 */
template <class T>
class matrixView {
  //! Matrix rows and cols
  size_t rows_, cols_;
  //! Pointer to matrix data
  T* data_;
public:
  //! Default constructor
  inline explicit matrixView() : rows_(0), cols_(0), data_((T*)0) {}
  //! Main constructor to set member data
  inline matrixView(size_t rows, size_t cols, T* data)
    : rows_(rows), cols_(cols), data_(data) {}
  //! Destructor
  inline ~matrixView() {}// nothing to delete because it's wrapper
  
  //! Assignement operator
  inline matrixView& operator= (const matrixView& mv)
  {
    rows_ = mv.rows_; cols_ = mv.cols_;
    data_ = mv.data_;
    return *this;
  }
  
  //! return the element of the ith row and jth colum of the underlying matrix
  inline const T& operator() (size_t i, size_t j)
  {
#ifdef CHECK_SIZE
    if ( data_ == 0 ) throw Error("matrixView<>::operator(), empty array");
    if ( !( (i >= 0) && (i < rows_) ) ) {
      cerr << "i = " << i << ", rows = " << rows_ << endl;
      throw Error("matrixView<>::operator(), first index out of bound");
    }
    if ( !( (j >= 0) && (j < cols_) ) ) {
      cerr << "j = " << j << ", cols = " << cols_ << endl;
      throw Error("matrixView<>::operator(), second index out of bound");
    }
#endif
    
    return *(data_+i*cols_+j);
  }
  
  //! return the ith row
  inline vectorView<T> operator [] (size_t i) const
  {
#ifdef CHECK_SIZE
    if ( data_ == 0 ) throw Error("matrixView<>::operator[], empty array");
    
    if ( !( (i >= 0) && (i < rows_) ) ) {
      cerr << "i = " << i << ", rows = " << rows_ << endl;
      throw Error("matrixView<>::operator[], index out of bound");
    }
#endif
    return vectorView<T>(cols_,data_+i*cols_);
  }

  //! return number of rows
  inline size_t rows() const { return rows_;}
  //! return number of colums
  inline size_t cols() const { return cols_;}
  
  //! return pointer to data of the underlying matrix
  inline T* begin() const { return data_;}
};


#endif
